package main

import (
	"bufio"
	"dehub"
	"errors"
	"flag"
	"fmt"
	"io"
	"os"
	"strings"

	"gopkg.in/src-d/go-git.v4/plumbing"
)

func exitErr(err error) {
	fmt.Fprintf(os.Stderr, "exiting: %v\n", err)
	os.Stderr.Sync()
	os.Stdout.Sync()
	os.Exit(1)
}

type subCmdCtx struct {
	repo func() *dehub.Repo
	flag *flag.FlagSet
	args []string
}

func (sctx subCmdCtx) flagParse() {
	if err := sctx.flag.Parse(sctx.args); err != nil {
		exitErr(err)
	}
}

type subCmd struct {
	name, descr string
	body        func(sctx subCmdCtx) error
}

var subCmds = []subCmd{
	{
		name:  "commit",
		descr: "commits staged changes to the head of the current branch",
		body: func(sctx subCmdCtx) error {
			msg := sctx.flag.String("msg", "", "Commit message to use")
			accountID := sctx.flag.String("account-id", "", "Account to sign commit as")
			sctx.flagParse()

			if *msg == "" || *accountID == "" {
				flag.PrintDefaults()
				return errors.New("-msg and -account-id are both required")
			}

			cfg, err := sctx.repo().LoadConfig()
			if err != nil {
				return err
			}

			var account dehub.Account
			var ok bool
			for _, account = range cfg.Accounts {
				if account.ID == *accountID {
					ok = true
					break
				}
			}
			if !ok {
				return fmt.Errorf("account ID %q not found in config", *accountID)
			} else if l := len(account.Signifiers); l == 0 || l > 1 {
				return fmt.Errorf("account %q has %d signifiers, only one is supported right now", *accountID, l)
			}

			sig := account.Signifiers[0]
			sigInt, err := sig.Interface()
			if err != nil {
				return fmt.Errorf("could not cast %+v to SignifierInterface: %w", sig, err)
			}

			tc, err := sctx.repo().NewTrunkCommit(*msg, *accountID, sigInt)
			if err != nil {
				return fmt.Errorf("could not construct trunk commit: %w", err)
			}

			hash, err := sctx.repo().Commit(tc, *accountID)
			if err != nil {
				return fmt.Errorf("could not commit trunk commit: %w", err)
			}

			fmt.Printf("changes committed to HEAD as %s\n", hash)
			return nil
		},
	},
	{
		name:  "verify",
		descr: "verifies one or more commits as having the proper credentials",
		body: func(sctx subCmdCtx) error {
			rev := sctx.flag.String("rev", "HEAD", "Revision of commit to verify")
			sctx.flagParse()

			h, err := sctx.repo().GitRepo.ResolveRevision(plumbing.Revision(*rev))
			if err != nil {
				return fmt.Errorf("could not resolve revision %q: %w", *rev, err)
			}

			if err := sctx.repo().VerifyTrunkCommit(*h); err != nil {
				return fmt.Errorf("could not verify commit at %q (%s): %w", *rev, *h, err)
			}

			fmt.Printf("commit at %q (%s) is good to go!\n", *rev, *h)
			return nil
		},
	},
	{
		name:  "hook",
		descr: "use dehub as a git hook",
		body: func(sctx subCmdCtx) error {
			preRcv := sctx.flag.Bool("pre-receive", false, "Use dehub as a server-side pre-receive hook")
			sctx.flagParse()

			if !*preRcv {
				flag.PrintDefaults()
				return errors.New("must set the hook type")
			}

			br := bufio.NewReader(os.Stdin)
			for {
				line, err := br.ReadString('\n')
				if errors.Is(err, io.EOF) {
					return nil
				} else if err != nil {
					return fmt.Errorf("error reading next line from stdin: %w", err)
				}

				lineParts := strings.Fields(line)
				if len(lineParts) < 3 {
					return fmt.Errorf("malformed pre-receive hook stdin line %q", line)
				}

				if plumbing.ReferenceName(lineParts[2]) != dehub.Trunk {
					return fmt.Errorf("only commits to the trunk branch are allowed at the moment (tried to push to %q)", lineParts[2])
				}

				// the zeroRevision gets sent on the very first push
				const zeroRevision plumbing.Revision = "0000000000000000000000000000000000000000"

				fromRev := plumbing.Revision(lineParts[0])
				var fromHash *plumbing.Hash
				if fromRev != zeroRevision {
					fromHash, err = sctx.repo().GitRepo.ResolveRevision(fromRev)
					if err != nil {
						return fmt.Errorf("unable to resolve revision %q: %w", fromRev, err)
					}
				}

				toRev := plumbing.Revision(lineParts[1])
				toHash, err := sctx.repo().GitRepo.ResolveRevision(toRev)
				if err != nil {
					return fmt.Errorf("unable to resolve revision %q: %w", toRev, err)
				}

				toCommit, err := sctx.repo().GitRepo.CommitObject(*toHash)
				if err != nil {
					return fmt.Errorf("unable to find commit %q: %w", *toHash, err)
				}

				var hashesToCheck []plumbing.Hash
				var found bool
				for currCommit := toCommit; ; {
					hashesToCheck = append(hashesToCheck, currCommit.Hash)
					if currCommit.NumParents() == 0 {
						break
					} else if currCommit.NumParents() > 1 {
						return fmt.Errorf("commit %q has more than one parent: %+v",
							currCommit.Hash, currCommit.ParentHashes)
					}

					parentCommit, err := currCommit.Parent(0)
					if err != nil {
						return fmt.Errorf("unable to get parent of commit %q: %w", currCommit.Hash, err)
					} else if fromHash != nil && parentCommit.Hash == *fromHash {
						found = true
						break
					}
					currCommit = parentCommit
				}
				if !found && fromHash != nil {
					return fmt.Errorf("unable to find commit %q as an ancestor of %q", *fromHash, *toHash)
				}

				for i := len(hashesToCheck) - 1; i >= 0; i-- {
					hash := hashesToCheck[i]
					fmt.Printf("Verifying trunk commit %q\n", hash)
					if err := sctx.repo().VerifyTrunkCommit(hash); err != nil {
						return fmt.Errorf("could not verify trunk commit %q", hash)
					}
				}
				fmt.Println("All pushed commits have been verified, well done.")
				return nil
			}
		},
	},
}

func main() {
	var pickedSubCmd bool
	var subCmd subCmd
	flagSet := flag.NewFlagSet("dehub", flag.ContinueOnError)
	flagSet.Usage = func() {
		cmd := "<command>"
		if pickedSubCmd {
			cmd = subCmd.name
		}
		fmt.Printf("USAGE: %s [global flags] %s [command flags]\n\n", os.Args[0], cmd)

		if !pickedSubCmd {
			fmt.Println("COMMANDS")
			for _, subCmd := range subCmds {
				fmt.Printf("\t%s : %s\n", subCmd.name, subCmd.descr)
			}
			fmt.Println("")
		}

		fmt.Println("GLOBAL FLAGS")
		flagSet.PrintDefaults()
		fmt.Println("")
	}

	bare := flagSet.Bool("bare", false, "If set then dehub will expect to be working with a bare repo")
	if err := flagSet.Parse(os.Args[1:]); err != nil {
		exitErr(err)
		return
	}

	args := flagSet.Args()
	if len(args) < 1 {
		flagSet.Usage()
		exitErr(errors.New("no command selected"))
	}

	subCmdName, args := strings.ToLower(args[0]), args[1:]
	for _, subCmd = range subCmds {
		if pickedSubCmd = subCmd.name == subCmdName; pickedSubCmd {
			break
		}
	}

	if !pickedSubCmd {
		flagSet.Usage()
		exitErr(fmt.Errorf("unknown command %q", subCmdName))
	}

	subFlagSet := flag.NewFlagSet(subCmd.name, flag.ExitOnError)
	subFlagSet.Usage = func() {
		flagSet.Usage()
		fmt.Println("COMMAND FLAGS")
		subFlagSet.PrintDefaults()
		fmt.Println("")
	}

	var r *dehub.Repo
	repoFn := func() *dehub.Repo {
		if r != nil {
			return r
		}
		var err error
		if r, err = dehub.OpenRepo(".", dehub.OpenBare(*bare)); err != nil {
			wd, _ := os.Getwd()
			err = fmt.Errorf("failed to OpenRepo at %q: %w", wd, err)
			exitErr(err)
		}
		return r
	}

	err := subCmd.body(subCmdCtx{
		repo: repoFn,
		flag: subFlagSet,
		args: args,
	})
	if err != nil {
		exitErr(err)
	}
}
